import inspect
import json
from libhoney.internal import json_default_handler

class FieldHolder:
    '''A FieldHolder is the generalized class that stores fields and dynamic
       fields. It should not be used directly; only through the subclasses'''

    def __init__(self):
        self._data = {}
        self._dyn_fields = set()

    def __add__(self, other):
        '''adding two field holders merges the data with other overriding
           any fields they have in common'''
        self._data.update(other._data)
        self._dyn_fields.update(other._dyn_fields)
        return self

    def __eq__(self, other):
        '''two FieldHolders are equal if their datasets are equal'''
        return ((self._data, self._dyn_fields) ==
                (other._data, other._dyn_fields))

    def __ne__(self, other):
        '''two FieldHolders are equal if their datasets are equal'''
        return not self.__eq__(other)

    def add_field(self, name, val):
        self._data[name] = val

    def add_dynamic_field(self, fn):
        if not inspect.isroutine(fn):
            raise TypeError("add_dynamic_field requires function argument")
        self._dyn_fields.add(fn)

    def add(self, data):
        try:
            for k, v in data.items():
                self.add_field(k, v)
        except AttributeError:
            raise TypeError("add requires a dict-like argument")

    def is_empty(self):
        '''returns true if there is no data in this FieldHolder'''
        return len(self._data) == 0

    def __str__(self):
        '''returns a JSON blob of the fields in this holder'''
        return json.dumps(self._data, default=json_default_handler)



def json_default_handler(obj):
    ''' this function handles values that the json encoder does not understand
    by attempting to call the object's __str__ method. '''
    try:
        return str(obj)
    except Exception:
        return 'libhoney was unable to encode value'


import datetime
import random
from contextlib import contextmanager

import libhoney.state as state
from libhoney.fields import FieldHolder


class Event(object):
    '''An Event is a collection of fields that will be sent to Honeycomb.'''

    def __init__(self, data={}, dyn_fields=[], fields=FieldHolder(), client=None):
        if client is None:
            client = state.G_CLIENT

        # copy configuration from client
        self.client = client
        if self.client:
            self.writekey = client.writekey
            self.dataset = client.dataset
            self.api_host = client.api_host
            self.sample_rate = client.sample_rate
        else:
            self.writekey = None
            self.dataset = None
            self.api_host = 'https://api.honeycomb.io'
            self.sample_rate = 1

        # populate the event's fields
        self._fields = FieldHolder()  # get an empty FH
        if self.client:
            self._fields += self.client.fields # fill it with the client fields
        self._fields.add(data)        # and anything passed in
        [self._fields.add_dynamic_field(fn) for fn in dyn_fields]
        self._fields += fields

        # fill in other info
        self.created_at = datetime.datetime.utcnow()
        self.metadata = None
        # execute all the dynamic functions and add their data
        for fn in self._fields._dyn_fields:
            self._fields.add_field(fn.__name__, fn())


    def add_field(self, name, val):
        self._fields.add_field(name, val)

    def add_metadata(self, md):
        '''Add metadata to an event. This metadata is handed back to you in
        the response queue. It is not transmitted to Honeycomb; it is a place
        for you to put identifying information to understand which event a
        response queue object represents.'''
        self.metadata = md

    def add(self, data):
        self._fields.add(data)

    @contextmanager
    def timer(self, name):
        '''timer is a context for timing (in milliseconds) a function call.

        Example:

            ev = Event()
            with ev.timer("database_dur_ms"):
                do_database_work()

           will add a field (name, duration) indicating how long it took to run
           do_database_work()'''
        start = datetime.datetime.now()
        yield
        duration = datetime.datetime.now() - start
        # report in ms
        self.add_field(name, duration.total_seconds() * 1000)

    def send(self):
        '''send queues this event for transmission to Honeycomb.

        Will drop sampled events when sample_rate > 1,
        and ensure that the Honeycomb datastore correctly considers it
        as representing `sample_rate` number of similar events.'''
        # warn if we're not using a client instance and global libhoney
        # is not initialized. This will result in a noop, but is better
        # than crashing the caller if they forget to initialize
        if self.client is None:
            state.warn_uninitialized()
            return

        if _should_drop(self.sample_rate):
            self.client.send_dropped_response(self)
            return

        self.send_presampled()

    def send_presampled(self):
        '''send_presampled queues this event for transmission to Honeycomb.

        Caller is responsible for sampling logic - will not drop any events
        for sampling. Defining a `sample_rate` will ensure that the Honeycomb
        datastore correctly considers it as representing `sample_rate` number
        of similar events.

        Raises SendError if no fields are defined or critical attributes not
        set (writekey, dataset, api_host).'''
        if self._fields.is_empty():
            self.client.log("No metrics added to event. Won't send empty event.")
            return
        if self.api_host == "":
            self.client.log("No api_host for Honeycomb. Can't send to the Great Unknown.")
            return
        if self.writekey == "":
            self.client.log("No writekey specified. Can't send event.")
            return
        if self.dataset == "":
            self.client.log(
                "No dataset for Honeycomb. Can't send event without knowing which dataset it belongs to.")
            return

        if self.client:
            self.client.send(self)
        else:
            state.warn_uninitialized()

    def __str__(self):
        return str(self._fields)

    def fields(self):
        return self._fields._data

def _should_drop(rate):
    '''returns true if the sample should be dropped'''
    return random.randint(1, rate) != 1


'''Tests for libhoney/__init__.py'''

import datetime
import json
import mock
import unittest

import libhoney


def sample_dyn_fn():
    return "dyna", "magic"


class FakeTransmitter():
    def __init__(self, id):
        self.id = id

    def start(self):
        pass

    def close(self):
        pass

    def get_response_queue(self):
        return None


class TestGlobalScope(unittest.TestCase):
    def setUp(self):
        # reset global state with each test
        libhoney.close()
        self.mock_xmit = mock.Mock(return_value=mock.Mock())

    def test_init(self):
        ft = FakeTransmitter(3)
        with mock.patch('libhoney.client.Transmission') as m_xmit:
            m_xmit.return_value = ft
            libhoney.init(writekey="wk", dataset="ds", sample_rate=3,
                          api_host="uuu", max_concurrent_batches=5,
                          block_on_response=True)
            self.assertEqual(libhoney.state.G_CLIENT.writekey, "wk")
            self.assertEqual(libhoney.state.G_CLIENT.dataset, "ds")
            self.assertEqual(libhoney.state.G_CLIENT.api_host, "uuu")
            self.assertEqual(libhoney.state.G_CLIENT.sample_rate, 3)
            self.assertEqual(libhoney.state.G_CLIENT.xmit, ft)
            self.assertEqual(libhoney.state.G_CLIENT._responses, None)
            m_xmit.assert_called_with(
                block_on_response=True, block_on_send=False,
                max_concurrent_batches=5, user_agent_addition='',
                debug=False,
            )

    def test_close(self):
        mock_client = mock.Mock()
        libhoney.state.G_CLIENT = mock_client
        libhoney.close()
        mock_client.close.assert_called_with()
        self.assertEqual(libhoney.state.G_CLIENT, None)

    def test_close_noop_if_not_init(self):
        """
        libhoney.close() should noop if never initialized
        """
        try:
            libhoney.close()
        except AttributeError:
            self.fail('libhoney threw an exception on '
                      'an uninitialized close.')

    def test_add_field(self):
        libhoney.init()
        ed = {"whomp": True}
        libhoney.add_field("whomp", True)
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, ed)

    def test_add_dynamic_field(self):
        libhoney.init()
        ed = set([sample_dyn_fn])
        libhoney.add_dynamic_field(sample_dyn_fn)
        self.assertEqual(libhoney.state.G_CLIENT.fields._dyn_fields, ed)

    def test_add(self):
        libhoney.init()
        ed = {"whomp": True}
        libhoney.add(ed)
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, ed)


class TestFieldHolder(unittest.TestCase):
    def setUp(self):
        # reset global state with each test
        libhoney.close()

    def test_add_field(self):
        libhoney.init()
        expected_data = {}
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, expected_data)
        self.assertTrue(libhoney.state.G_CLIENT.fields.is_empty())
        libhoney.add_field("foo", 4)
        expected_data["foo"] = 4
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, expected_data)
        self.assertFalse(libhoney.state.G_CLIENT.fields.is_empty())
        libhoney.add_field("bar", "baz")
        expected_data["bar"] = "baz"
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, expected_data)
        libhoney.add_field("foo", 6)
        expected_data["foo"] = 6
        self.assertEqual(libhoney.state.G_CLIENT.fields._data, expected_data)

    def test_add_dynamic_field(self):
        libhoney.init()
        expected_dyn_fns = set()
        self.assertEqual(
            libhoney.state.G_CLIENT.fields._dyn_fields, expected_dyn_fns)
        libhoney.add_dynamic_field(sample_dyn_fn)
        expected_dyn_fns.add(sample_dyn_fn)
        self.assertEqual(
            libhoney.state.G_CLIENT.fields._dyn_fields, expected_dyn_fns)
        # adding a second time should still only have one element
        libhoney.add_dynamic_field(sample_dyn_fn)
        self.assertEqual(
            libhoney.state.G_CLIENT.fields._dyn_fields, expected_dyn_fns)
        with self.assertRaises(TypeError):
            libhoney.add_dynamic_field("foo")


class TestBuilder(unittest.TestCase):
    def setUp(self):
        # reset global state with each test
        libhoney.close()

    def test_new_builder(self):
        libhoney.init()
        # new builder, no arguments
        b = libhoney.Builder()
        self.assertEqual(b._fields._data, {})
        self.assertEqual(b._fields._dyn_fields, set())
        # new builder, passed in data and dynfields
        expected_data = {"aa": 1}
        expected_dyn_fns = set([sample_dyn_fn])
        b = libhoney.Builder(expected_data, expected_dyn_fns)
        self.assertEqual(b._fields._data, expected_data)
        self.assertEqual(b._fields._dyn_fields, expected_dyn_fns)
        # new builder, inherited data and dyn_fields
        libhoney.state.G_CLIENT.fields._data = expected_data
        libhoney.state.G_CLIENT.fields._dyn_fields = expected_dyn_fns
        b = libhoney.Builder()
        self.assertEqual(b._fields._data, expected_data)
        self.assertEqual(b._fields._dyn_fields, expected_dyn_fns)
        # new builder, merge inherited data and dyn_fields and arguments

        def sample_dyn_fn2():
            return 5
        expected_data = {"aa": 1, "b": 2}
        expected_dyn_fns = set([sample_dyn_fn, sample_dyn_fn2])
        b = libhoney.Builder({"b": 2}, [sample_dyn_fn2])
        self.assertEqual(b._fields._data, expected_data)
        self.assertEqual(b._fields._dyn_fields, expected_dyn_fns)

    def test_add_field(self):
        libhoney.init()
        b = libhoney.Builder()
        expected_data = {}
        self.assertEqual(b._fields._data, expected_data)
        b.add_field("foo", 4)
        expected_data["foo"] = 4
        self.assertEqual(b._fields._data, expected_data)
        b.add_field("bar", "baz")
        expected_data["bar"] = "baz"
        self.assertEqual(b._fields._data, expected_data)
        b.add_field("foo", 6)
        expected_data["foo"] = 6
        self.assertEqual(b._fields._data, expected_data)

    def test_add_dynamic_field(self):
        libhoney.init()
        b = libhoney.Builder()
        expected_dyn_fns = set()
        self.assertEqual(b._fields._dyn_fields, expected_dyn_fns)
        b.add_dynamic_field(sample_dyn_fn)
        expected_dyn_fns.add(sample_dyn_fn)
        self.assertEqual(b._fields._dyn_fields, expected_dyn_fns)
        with self.assertRaises(TypeError):
            b.add_dynamic_field("foo")

    def test_add(self):
        libhoney.init()
        b = libhoney.Builder()
        expected_data = {"a": 1, "b": 3}
        b.add(expected_data)
        self.assertEqual(b._fields._data, expected_data)
        expected_data.update({"c": 3, "d": 4})
        b.add({"c": 3, "d": 4})
        self.assertEqual(b._fields._data, expected_data)

    def test_new_event(self):
        libhoney.init()
        b = libhoney.Builder()
        b.sample_rate = 5
        expected_data = {"a": 1, "b": 3}
        b.add(expected_data)
        ev = b.new_event()
        self.assertEqual(b._fields, ev._fields)
        self.assertEqual(ev.sample_rate, 5)
        ev.add_field("3", "c")
        self.assertNotEqual(b._fields, ev._fields)
        # move to event testing when written
        self.assertEqual(json.loads(str(ev)), {"a": 1, "3": "c", "b": 3})

    def test_clone_builder(self):
        libhoney.init()

        b = libhoney.Builder()
        b.dataset = "newds"
        b.add_field("e", 9)
        b.add_dynamic_field(sample_dyn_fn)
        c = b.clone()
        self.assertEqual(b._fields, c._fields)
        c.add_field("f", 10)
        b.add_field("g", 11)
        self.assertEqual(b._fields._data, {"e": 9, "g": 11})
        self.assertEqual(c._fields._data, {"e": 9, "f": 10})
        self.assertEqual(c.dataset, "newds")


class TestEvent(unittest.TestCase):
    def setUp(self):
        # reset global state with each test
        libhoney.close()

    def test_timer(self):
        libhoney.init()

        class fakeDate:
            def setNow(self, time):
                self.time = time

            def now(self):
                return self.time

            def utcnow(self):
                return self.time

        with mock.patch('libhoney.event.datetime') as m_datetime:
            fakeStart = datetime.datetime(2016, 1, 2, 3, 4, 5, 6)
            fakeEnd = fakeStart + datetime.timedelta(milliseconds=5)
            fd = fakeDate()
            fd.setNow(fakeStart)
            m_datetime.datetime = fd
            ev = libhoney.Event()
            with ev.timer("howlong"):
                fd.setNow(fakeEnd)
            self.assertEqual(ev._fields._data, {"howlong": 5})
            self.assertEqual(ev.created_at, fakeStart)

    def test_str(self):
        libhoney.init()
        ev = libhoney.Event()
        ev.add_field("obj", {"a": 1})
        ev.add_field("string", "a:1")
        ev.add_field("number", 5)
        ev.add_field("boolean", True)
        ev.add_field("null", None)

        serialized = str(ev)
        self.assertTrue('"obj": {"a": 1}' in serialized)
        self.assertTrue('"string": "a:1"' in serialized)
        self.assertTrue('"number": 5' in serialized)
        self.assertTrue('"boolean": true' in serialized)
        self.assertTrue('"null": null' in serialized)

    def test_send(self):
        with mock.patch('libhoney.client.Transmission') as m_xmit:
            libhoney.init()
            ev = libhoney.Event()
            # override inherited api_host from client
            ev.api_host = ""
            ev.add_field("f", "g")
            ev.api_host = "myhost"
            ev.writekey = "letmewrite"
            ev.dataset = "storeme"
            ev.send()
            m_xmit.return_value.send.assert_called_with(ev)

    def test_send_sampling(self):
        with mock.patch('libhoney.client.Transmission') as m_xmit,\
                mock.patch('libhoney.event._should_drop') as m_sd:
            m_sd.return_value = True
            libhoney.init(writekey="wk", dataset="ds")

            # test that send() drops when should_drop is true
            ev = libhoney.Event()
            ev.add_field("foo", 1)
            ev.send()
            m_xmit.return_value.send.assert_not_called()
            m_sd.assert_called_with(1)
            ev = libhoney.Event()
            ev.add_field("foo", 1)
            ev.sample_rate = 5
            ev.send()
            m_xmit.return_value.send.assert_not_called()
            m_sd.assert_called_with(5)

            # and actually sends them along when should_drop is false
            m_sd.reset_mock()
            m_xmit.reset_mock()
            m_sd.return_value = False

            ev = libhoney.Event()
            ev.add_field("f", "g")
            ev.api_host = "myhost"
            ev.writekey = "letmewrite"
            ev.dataset = "storeme"
            ev.send()
            m_xmit.return_value.send.assert_called_with(ev)
            m_sd.assert_called_with(1)
            ev.sample_rate = 5
            ev.send()
            m_xmit.return_value.send.assert_called_with(ev)
            m_sd.assert_called_with(5)

            # test that send_presampled() does not drop
            m_sd.reset_mock()
            m_xmit.reset_mock()
            ev.send_presampled()
            m_xmit.return_value.send.assert_called_with(ev)
            m_sd.assert_not_called()

            m_sd.reset_mock()
            m_xmit.reset_mock()
            ev.sample_rate = 5
            ev.send_presampled()
            m_xmit.return_value.send.assert_called_with(ev)
            m_sd.assert_not_called()



'''Tests for libhoney/client.py'''

import unittest
import mock

import libhoney
import libhoney.client as client

def sample_dyn_fn():
    return "dyna", "magic"


class TestClient(unittest.TestCase):
    def setUp(self):
        libhoney.close()

        self.tx = mock.Mock()
        self.m_xmit = mock.patch('libhoney.client.Transmission')
        self.m_xmit.start().return_value = self.tx

    def tearDown(self):
        self.m_xmit.stop()

    def test_init(self):
        c = client.Client(writekey="foo", dataset="bar", api_host="blup",
                          sample_rate=2)
        self.assertEqual(c.writekey, "foo")
        self.assertEqual(c.dataset, "bar")
        self.assertEqual(c.api_host, "blup")
        self.assertEqual(c.sample_rate, 2)
        self.assertEqual(c.xmit, self.tx)

    def test_close(self):
        c = client.Client(writekey="foo", dataset="bar", api_host="blup",
                          sample_rate=2)
        c.close()
        self.tx.close.assert_called_with()
        self.assertEqual(c.xmit, None)

    def test_close_noop_if_not_init(self):
        """
        libhoney.close() should noop if never initialized
        """
        try:
            c = client.Client()
            c.close()
            # close again, should be a noop and not cause an exception
            c.close()
        except AttributeError:
            self.fail('libhoney threw an exception on '
                           'an uninitialized close.')

    def test_add_field(self):
        ed = {"whomp": True}
        with client.Client() as c:
            c.add_field("whomp", True)
            self.assertEqual(c.fields._data, ed)

    def test_add_dynamic_field(self):
        ed = set([sample_dyn_fn])
        with client.Client() as c:
            c.add_dynamic_field(sample_dyn_fn)
            self.assertEqual(c.fields._dyn_fields, ed)

    def test_add(self):
        ed = {"whomp": True}
        with client.Client() as c:
            c.add(ed)
            self.assertEqual(c.fields._data, ed)

    def test_new_event(self):
        with client.Client(writekey="client_key", dataset="client_dataset") as c:
            c.add_field("whomp", True)
            c.add_dynamic_field(sample_dyn_fn)

            ev = c.new_event({"field1": 1})
            ev.add_field("field2", 2)
            # ensure client config is passed through
            self.assertEqual(ev.client, c)
            self.assertEqual(ev.writekey, "client_key")
            self.assertEqual(ev.dataset, "client_dataset")
            # ensure client fields are passed on
            self.assertEqual(ev._fields._data, {"field1": 1, "field2": 2, "whomp": True, "sample_dyn_fn": ("dyna", "magic")})
            self.assertEqual(ev._fields._dyn_fields, set([sample_dyn_fn]))

    def test_new_builder(self):
        with client.Client(writekey="client_key", dataset="client_dataset") as c:
            c.add_field("whomp", True)
            c.add_dynamic_field(sample_dyn_fn)

            b = c.new_builder()
            # ensure client config is passed through
            self.assertEqual(b.client, c)
            self.assertEqual(b.writekey, "client_key")
            self.assertEqual(b.dataset, "client_dataset")
            b.add_field("builder_field", 1)

            ev = b.new_event()
            # ensure client config gets passed on
            self.assertEqual(ev.client, c)
            self.assertEqual(ev.writekey, "client_key")
            self.assertEqual(ev.dataset, "client_dataset")
            ev.add_field("event_field", 2)
            self.assertEqual(ev.client, c)
            # ensure client and builder fields are passed on
            self.assertEqual(ev._fields._data, {"builder_field": 1, "event_field": 2, "whomp": True, "sample_dyn_fn": ("dyna", "magic")})
            self.assertEqual(ev._fields._dyn_fields, set([sample_dyn_fn]))

    def test_send(self):
        ''' ensure that Event's `send()` calls the client `send()` method '''
        with client.Client(writekey="mykey", dataset="something") as c:
            # explicitly use a different object for Transmission than is
            # defined in setUp, to ensure we aren't using the global
            # xmit in libhoney
            c.xmit = mock.Mock()

            ev = c.new_event()
            ev.add_field("foo", "bar")
            ev.send()
            c.xmit.send.assert_called_with(ev)

    def test_send_global(self):
        ''' ensure that events sent using the global libhoney config work '''
        with client.Client(writekey="mykey", dataset="something") as c:
            # explicitly use a different object for Transmission than is
            # defined in setUp, to ensure we aren't using the global
            # xmit in libhoney
            c.xmit = mock.Mock()

            libhoney.init(writekey="someotherkey", dataset="somethingelse")
            ev = libhoney.Event()
            self.assertEqual(ev.writekey, "someotherkey")
            self.assertEqual(ev.dataset, "somethingelse")
            ev.add_field("global", "event")
            ev.send()
            # test our assumption about what's actually mocked
            self.assertEqual(libhoney.state.G_CLIENT.xmit, self.tx)
            # check that we used the global xmit
            self.tx.send.assert_called_with(ev)
            # check that the client xmit was not used
            self.assertFalse(c.xmit.send.called)

    def test_send_dropped_response(self):
        with mock.patch('libhoney.event._should_drop') as m_drop:
            m_drop.return_value = True

            with client.Client(writekey="mykey", dataset="something") as c:
                ev = c.new_event()
                ev.add_field("a", "b")
                ev.send()

                c.responses().put_nowait.assert_called_with({
                    "status_code": 0,
                    "duration": 0,
                    "metadata": ev.metadata,
                    "body": "",
                    "error": "event dropped due to sampling",
                })

    def test_xmit_override(self):
        '''verify that the client accepts an alternative Transmission'''
        mock_xmit = mock.Mock()
        with client.Client(transmission_impl=mock_xmit) as c:
            self.assertEqual(c.xmit, mock_xmit)

class TestClientFlush(unittest.TestCase):
    ''' separate test class because we don't want to mock transmission'''
    def test_flush(self):
        mock_xmit = mock.Mock(spec=libhoney.transmission.Transmission)
        with client.Client(transmission_impl=mock_xmit) as c:
            # start gets called when the class is initialized
            mock_xmit.start.assert_called_once_with()
            mock_xmit.reset_mock()
            c.flush()

            mock_xmit.close.assert_called_once_with()
            mock_xmit.start.assert_called_once_with()

        mock_xmit = mock.Mock(spec=libhoney.transmission.TornadoTransmission)
        with client.Client(transmission_impl=mock_xmit) as c:
            # start gets called when the class is initialized
            mock_xmit.start.assert_called_once_with()
            mock_xmit.reset_mock()
            c.flush()

            # we don't call close/start on TornadoTransmission because we can't
            # force a flush in an async environment.
            mock_xmit.close.assert_not_called()
            mock_xmit.start.assert_not_called()



import sys
from setuptools import setup
sys.path.append('libhoney/')
from version import VERSION


setup(name='libhoney',
      version=VERSION,
      description='Python library for sending data to Honeycomb',
      url='https://github.com/honeycombio/libhoney-py',
      author='Honeycomb.io',
      author_email='feedback@honeycomb.io',
      license='Apache',
      packages=['libhoney'],
      install_requires=[
          'requests',
          'statsd',
          'six',
          'futures; python_version == "2.7"'
      ],
      tests_require=[
        'mock',
        'pbr',
        'requests-mock',
        'bumpversion',
        'tornado>=5.0,<6.0',
      ],
      test_suite='libhoney',
      zip_safe=False)



'''Tests for libhoney/transmission.py'''

import datetime
import unittest

import mock
import six
import tornado
import libhoney
import transmission

PATCH_NAMESPACE='transmission'
if six.PY2:
    PATCH_NAMESPACE='libhoney.transmission'


class TestTornadoTransmissionInit(unittest.TestCase):
    def test_defaults(self):
        t = transmission.TornadoTransmission()
        self.assertIsInstance(t.batch_sem, tornado.locks.Semaphore)
        self.assertIsInstance(t.pending, tornado.queues.Queue)
        self.assertIsInstance(t.responses, tornado.queues.Queue)
        self.assertEqual(t.block_on_send, False)
        self.assertEqual(t.block_on_response, False)

    def test_args(self):
        t = transmission.TornadoTransmission(max_concurrent_batches=4, block_on_send=True, block_on_response=True)
        t.start()
        self.assertEqual(t.block_on_send, True)
        self.assertEqual(t.block_on_response, True)
        t.close()

    def test_user_agent_addition(self):
        ''' ensure user_agent_addition is included in the User-Agent header '''
        with mock.patch(PATCH_NAMESPACE + '.AsyncHTTPClient') as m_client:
            transmission.TornadoTransmission(user_agent_addition='foo/1.0')
            expected = "libhoney-py/" + libhoney.version.VERSION + " foo/1.0"
            m_client.assert_called_once_with(
                force_instance=True,
                defaults=dict(user_agent=expected),
            )


class TestTornadoTransmissionSend(unittest.TestCase):
    def test_send(self):
        with mock.patch(PATCH_NAMESPACE+'.AsyncHTTPClient') as m_http,\
                mock.patch('statsd.StatsClient') as m_statsd:
            m_http.return_value = mock.Mock()
            m_statsd.return_value = mock.Mock()
            @tornado.gen.coroutine
            def _test():
                t = transmission.TornadoTransmission()
                t.start()

                ev = mock.Mock(metadata=None, writekey="abc123",
                               dataset="blargh", api_host="https://example.com",
                               sample_rate=1, created_at=datetime.datetime.now())
                ev.fields.return_value = {"foo": "bar"}
                t.send(ev)

                # wait on the batch to be "sent"
                # we can detect this when data has been inserted into the
                # batch data dictionary
                while not t.batch_data:
                    yield tornado.gen.sleep(0.01)
                t.close()

            tornado.ioloop.IOLoop.current().run_sync(_test)
            m_statsd.return_value.incr.assert_any_call("messages_queued")
            self.assertTrue(m_http.return_value.fetch.called)


class TestTornadoTransmissionQueueOverflow(unittest.TestCase):
     def test_send(self):
         with mock.patch('statsd.StatsClient') as m_statsd:
            m_statsd.return_value = mock.Mock()

            t = transmission.TornadoTransmission()
            t.pending = tornado.queues.Queue(maxsize=2)
            t.responses = tornado.queues.Queue(maxsize=1)
            # we don't call start on transmission here, which will cause
            # the queue to pile up

            t.send(mock.Mock())
            t.send(mock.Mock())
            t.send(mock.Mock()) # should overflow sending and land on response
            m_statsd.return_value.incr.assert_any_call("queue_overflow")
            t.send(mock.Mock()) # shouldn't throw exception when response is full



class SendError(Exception):
    ''' raised when send is called on an event that cannot be sent, such as:
        - when it lacks a writekey
        - dataset is not specified
        - no fields are set
    '''
    pass
